/*

Copyright (C) 2011 Jarno Rajahalme

This is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the
Free Software Foundation; either version 2.1 of the License, or (at
your option) any later version.

This software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this software; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.


Wrapper for Apple libBLAS.dylib and libLAPACK.dylib

At least on the versions of OSX 10.6 so far (up and including 10.6.6)
these libraries are incompatible with 64 bit builds, as some functions
in libBLAS.dylib are not conforming to F2C calling conventions, as
they should.  This breaks them in 64-bit builds on the x86_64
architecture.

Newer gfortran compoilers no longer default to the F2C calling
convention.  These wrappers map the F2C conformant functions in
libBLAS and libLAPACK to the native gfortran calling convention, so
that the libraries can be used with software built for x86_64
architecture.

The wrappers are about as efficient as possible, given the task at
hand.  Internal calls to the wrapped functions within the wrapped
libraries are not affected in any way.
 
This file should be compiled with the following command:

$ gcc-4.2 -c -O3 -m64 -fomit-frame-pointer -W -Wall blaswrap.c

And linked like this (here assuming /opt/local/lib as the installation
location):

$ sudo ld -arch x86_64 -dylib -x blaswrap.o -o /opt/local/lib/libBLASWRAP.dylib -init _initVecLibWrappers -reexport_framework vecLib -lSystem

In addition to the wrapped functions, the line above makes the
resulting libBLASWRAP.dylib reexport all vecLib functions (including
libBLAS and libLAPACK), so that instead of e.g. "-framework vecLib",
or "-framework Accelerate", the users should link with
"-L/opt/local/lib -lBLASWRAP"

This wrapper has been successfully tested with GNU Octave 3.3.90,
configured with the option --with-blas="-lBLASWRAP",
as well as with the arpack library on a x86_64 OSX 10.6.6, using
GCC 4.5.

 */

#ifndef __x86_64__
#error blaswrap.c should only be compiled for 64-bit Intel targets!
#else

#include <dlfcn.h>

void abort (void);

/*
 * Apple LAPACK follows F2C calling convention,
 * Convert to normal gfortran calling convention
 */

static void (*f2c_lapack_func[])(void); // forward declaration for the wrapper

/*
 * LAPACK Wrapper, only need to convert the return value from double to float
 *
 * NOTE: This can only handle calls that have 6 or less arguments
 * Extra arguments have to be passed via stack, and separate stubs are needed
 * for each such call. (see below)
 */

typedef double (*F2C_CALL_7)(void *a1, void *a2, void *a3, void *a4, void *a5, void *a6, void *a7);
typedef double (*F2C_CALL_8)(void *a1, void *a2, void *a3, void *a4, void *a5, void *a6, void *a7, void *a8);

// Testing special LAPACK wrapper

#define F2C_LAPACK_CALL_8(name) \
  float name (void *a1, void *a2, void *a3, void *a4, void *a5, void *a6, void *a7, void *a8) \
  {									\
    return ((F2C_CALL_8)f2c_lapack_func[f2c_ ## name]) (a1, a2, a3, a4, a5, a6, a7, a8); \
  }

#define F2C_LAPACK_CALL_7(name) \
  float name (void *a1, void *a2, void *a3, void *a4, void *a5, void *a6, void *a7) \
  {									\
    return ((F2C_CALL_7)f2c_lapack_func[f2c_ ## name]) (a1, a2, a3, a4, a5, a6, a7); \
  }

#define F2C_LAPACK_CALL_0(name)			\
  void name (void) {					\
    __asm__ ("call	*%0\n\t"			\
	     "cvtsd2ss	%%xmm0,%%xmm0"			\
	     : : "m" (f2c_lapack_func[f2c_ ## name]) );	\
  }

#define F2C_LAPACK_CALL_1(name)	F2C_LAPACK_CALL_0(name)	 
#define F2C_LAPACK_CALL_2(name)	F2C_LAPACK_CALL_0(name)	 
#define F2C_LAPACK_CALL_3(name)	F2C_LAPACK_CALL_0(name)	 
#define F2C_LAPACK_CALL_4(name)	F2C_LAPACK_CALL_0(name)	 
#define F2C_LAPACK_CALL_5(name)	F2C_LAPACK_CALL_0(name)	 
#define F2C_LAPACK_CALL_6(name)	F2C_LAPACK_CALL_0(name)	 


#define F2C_LAPACK_CALL(name, args) F2C_LAPACK_CALL_ ## args (name)

#define ENUM_ITEM(name, args)			\
  f2c_ ## name, 

#define NAME_TO_STRING_CASE(name, args)		\
  case f2c_ ## name: return #name;

#define DEFINE_LAPACK_ENUM(name, list)	\
  typedef enum {			\
    list(ENUM_ITEM)			\
  } name;				\
  static const char*			\
  f2c_ ## name ## _name (name n) {	\
    switch (n) {			\
      list(NAME_TO_STRING_CASE)		\
    default: return "";			\
    }					\
  }					\
  list(F2C_LAPACK_CALL)

#define DEFINE_BLAS_ENUM(name, list)	\
  typedef enum {			\
    list(ENUM_ITEM)			\
  } name;				\
  static const char*			\
  f2c_ ## name ## _name(name n) {	\
    switch (n) {			\
      list(NAME_TO_STRING_CASE)		\
    default: return "";			\
    }					\
  }

/*
 * Lapack functions that need the return value converted from double to float
 * Simple stub is used for functions with all (0-6) args in registers,
 * individual stub types are used for longer argument counts.
 */
#define LAPACK_LIST(_)	\
  _(clangb_,7)		\
  _(clange_,6)		\
  _(clangt_,5)		\
  _(clanhb_,7)		\
  _(clanhe_,6)		\
  _(clanhp_,5)		\
  _(clanhs_,5)		\
  _(clanht_,4)		\
  _(clansb_,7)		\
  _(clansp_,5)		\
  _(clansy_,6)		\
  _(clantb_,8)		\
  _(clantp_,6)		\
  _(clantr_,8)		\
  _(scsum1_,3)		\
  _(second_,0)		\
  _(slamc3_,2)		\
  _(slamch_,1)		\
  _(slangb_,7)		\
  _(slange_,6)		\
  _(slangt_,5)		\
  _(slanhs_,5)		\
  _(slansb_,7)		\
  _(slansp_,5)		\
  _(slanst_,4)		\
  _(slansy_,6)		\
  _(slantb_,8)		\
  _(slantp_,6)		\
  _(slantr_,8)		\
  _(slapy2_,2)		\
  _(slapy3_,3)		\
  _(LAPACK_COUNT,0)
// XXX: f2c_LAPACK_COUNT should not be exported (e.g. ld -unexported_symbol _f2c_LAPACK_COUNT)

/*
 * These need more complex wrappers, as arguments need to be handled
 */
#define BLAS_LIST(_)	\
  _(cdotu_,6)		\
  _(zdotu_,6)		\
  _(cdotc_,6)		\
  _(zdotc_,6)		\
  _(BLAS_COUNT,0)

DEFINE_BLAS_ENUM(blas, BLAS_LIST)
DEFINE_LAPACK_ENUM(lapack, LAPACK_LIST)

/*
 * Function pointer arrays, indexed by the enums
 */
static void (*f2c_blas_func[f2c_BLAS_COUNT])(void) = { 0 };
static void (*f2c_lapack_func[f2c_LAPACK_COUNT])(void) = { 0 };

/*
 * Initialization: This is called by ld before any of the wrappers are called
 * Get the function pointers to the wrapped functions in Apple vecLib
 */
void initVecLibWrappers (void)
{
  void *apple_vecLib = dlopen ("/System/Library/Frameworks/vecLib.framework/Versions/A/vecLib", RTLD_LOCAL | RTLD_NOLOAD | RTLD_FIRST);
  if (0 == apple_vecLib)
    abort ();

  int i;
  for (i = 0; i < f2c_LAPACK_COUNT; i++)
    if (0 == (f2c_lapack_func[i] = dlsym(apple_vecLib, f2c_lapack_name(i))))
      abort ();  
  for (i = 0; i < f2c_BLAS_COUNT; i++)
    if (0 == (f2c_blas_func[i] = dlsym(apple_vecLib, f2c_blas_name(i))))
      abort ();  
}

/*
 * BLAS wrappers, F2C convention passes retuned complex as an extra first
 * argument
 */
typedef struct { float r, i; } complex;
typedef struct { double r, i; } doublecomplex;

typedef void (*F2C_CALL_6)(void *c, void *a1, void *a2, void *a3, void *a4, void *a5);

#define F2C_BLAS_CALL(type, name) \
type name (void *a1, void *a2, void *a3, void *a4, void *a5) \
{ \
  type cplx; \
  ((F2C_CALL_6)f2c_blas_func[f2c_ ## name]) (&cplx, a1, a2, a3, a4, a5); \
  return cplx; \
}

F2C_BLAS_CALL(complex, cdotu_)
F2C_BLAS_CALL(doublecomplex, zdotu_)
F2C_BLAS_CALL(complex, cdotc_)
F2C_BLAS_CALL(doublecomplex, zdotc_)

#endif
